
package ps6.graph;

/**
 * <code>Edge</code> represents a directed labeled edge in a graph.
 * <code>Edge</code> is immutable.
 * <p/>
 * Though the source, sink, and label of this edge may not be reassigned
 * through this interface, these elements may be references to the objects
 * they represent in the graph, which would potentially allow the client
 * to mutate an object in the graph.
 * When this is the case, these elements may not be mutated in such a way
 * that affects equals comparisons while they are elements in the graph.
 * (This is analogous to the restrictions on mutating keys in
 * the {@link java.util.Map} interface.)
 *
 * The generic parameter T is the type of Nodes which the edge connects
 * 
 * @specfield T source // the source of the edge
 * @specfield T sink // the sink of the edge
 * @specfield Object label // the label of the edge
 * 
 * @see java.util.Map
 * @see ps6.graph.DirectedLabeledMultiGraph
 * 
 * @author mbolin
 * @author cwo (modified for Java 1.5) 
 */
public interface Edge<T> {

	/**
	 * Returns the source of this edge.
	 * @returns source
	 */
	public T getSource();

	/**
	 * Returns the sink of this edge.
	 * @returns sink
	 */
	public T getSink();
	
	/**
	 * Returns the label of this edge.
	 * @returns label
	 */
	public Object getLabel();

	/**
	 * Compares the specified object with this edge for equality.
	 * 
	 * @returns <code>true</code> if <code>o</code> is an
	 * <code>Edge</code> and the <code>source</code>, <code>sink</code>, and <code>label</code>
	 * of this and <code>o</code> are <i>equal</i>.
	 * (Two elements <code>e1</code> and <code>e2</code>
	 * are <i>equal</i> if <code>(e1==null ? e2==null : e1.<b>equals</b>(e2))</code>.)
	 * 
	 * @see Object#hashCode()
	 * @see Object#equals(Object)
	 * @see #hashCode()
	 * @see ps6.graph.GraphUtils#eq(Object,Object)
	 */	
	public boolean equals(Object o);

	/**
     * Returns the hash code value for this edge. The hash code of an edge
     * is defined to be the result of the following calculation:
     * <pre>
     *  hashCode = 17;
     *  hashCode = 31*hashCode + ((getSource()==null) ? 19 : getSource().hashCode());
     *  hashCode = 31*hashCode + ((getLabel()==null)  ? 19 : getLabel().hashCode());
     *  hashCode = 31*hashCode + ((getSink()==null)   ? 19 : getSink().hashCode());
     * </pre>
     * This ensures that <code>edge1.equals(edge2)</code> implies that
     * <code>edge1.hashCode()==edge2.hashCode()</code> for any two edges,
     * <code>edge1</code> and <code>edge2</code>, as required by the general
     * contract of <code>Object.hashCode</code>.
	 * 
     * @returns the hash code value for this list.
     * 
     * @see Object#hashCode()
     * @see Object#equals(Object)
     * @see #equals(Object)
     * @see ps6.graph.GraphUtils#hcode(Object)
	 */
	public int hashCode();

}
