package ps6.graph;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;



/**
 * <code>EdgeSet</code> is a set of DefaultEdge objects that
 * is optimized for lookup and removal.
 *
 * The generic parameter T is the type of node that edges within this set
 * will connect
 *
 * @author mbolin
 * @author cwo (modified for Java 1.5) 
 */
public class EdgeSet<T> {

    /** all the incoming/outgoing edges from the node */
    private Set<Edge<T> > edges;
        
    /**
     * a mapping from labels to
     *   a mapping from nodes to edges
     * 
     */ 
    private Map<Object,Map<T, Edge<T> > > labelMap;
    
    private T node;
    
    private boolean incomingEdges;
    
    public EdgeSet(T node, boolean incomingEdges) {
        this.edges = new HashSet<Edge<T> >(); 
        this.labelMap = new HashMap<Object, Map<T, Edge<T> > >(); 
        this.node = node;
        this.incomingEdges = incomingEdges;
    }
    
    /**
     * @return true if the DefaultEdge is "anchored" to the node for this EdgeSet
     */
    private boolean testAnchor(Edge<T> e) {
        T oppositeNode = getOppositeNode(e);
        return GraphUtils.eq(node, oppositeNode);
    }
    
    private T getOppositeNode(Edge<T> e) {
        return (incomingEdges) ? e.getSource() : e.getSink(); 
    }
    
    public boolean add(Edge<T> e) {
        if (!testAnchor(e)) { return false; }
        boolean added = edges.add(e);
        if (added) {
            Map<T, Edge<T> > map = labelMap.get(e.getLabel());
            if (map == null) {
                map = new HashMap<T, Edge<T> >();
                labelMap.put(e.getLabel(), map);
            }
            map.put(getOppositeNode(e), e);
        }
        return added;
    }

    public boolean addAll(Collection<Edge<T> > c) {
        boolean changed = false;
        for (Iterator<Edge<T> > i = c.iterator(); i.hasNext(); ) {
            changed = changed || add(i.next());
        }
        return changed;
    }
    
    public void clear() {
        edges.clear();
        labelMap.clear();
    }

    public boolean contains(Edge<T> o) {
        return edges.contains(o);
    }

    public boolean containsAll(Collection<Edge<T> > c) {
        return edges.containsAll(c);
    }

    public boolean equals(Edge<T> obj) {
        return edges.equals(obj);
    }

    public int hashCode() {
        return edges.hashCode();
    }

    public boolean isEmpty() {
        return edges.isEmpty();
    }

    public Iterator<Edge<T> > iterator() {
        return edges.iterator();
    }

//  TODO make sure all methods after this point do the right thing
    
    public boolean remove(Edge<T> o) {
        return edges.remove(o);
    }

    public boolean removeAll(Collection<Edge<T> > c) {
        return edges.removeAll(c);
    }

    public boolean retainAll(Collection<Edge<T> > c) {
        return edges.retainAll(c);
    }

    public int size() {
        return edges.size();
    }

    public Object[] toArray() {
        return (Object[]) edges.toArray();
    }

    public Edge[] toArray(Edge[] a) {
        return edges.toArray(a);
    }

    public String toString() {
        return edges.toString();
    }
}
