package ps6.graph;

import java.util.Iterator;

/**
 * <code>GraphUtils</code> is a utility class for graphs in the
 * <code>ps6.graph</code> package.
 * 
 * @author mbolin
 * @author cwo (modified for Java 1.5)
 */
public final class GraphUtils {

	/** this class should not be instantiated */
	private GraphUtils() {
	}

	/**
	 * Compares two elements for equality.
	 * 
	 * This performs the more general test for equality used throughout this
	 * package where <code>o1.equals(o2)</code> is taken to mean
	 * <code>(e1==null ? e2==null : e1.<b>equals</b>(e2))</code> where the
	 * second <code><b>equals</b></code> is the method that tests for object
	 * equality.
	 * 
	 * @returns <code>true</code> if these two elements should be considered
	 *          equal
	 */
	public static boolean eq(Object o1, Object o2) {
		return ((o1 == null) ? o1 == o2 : o1.equals(o2));
	}

	/**
	 * Returns a hash code for this element.
	 * 
	 * @returns the hash code for the object, or 19 if <code>o</code> is
	 *          <code>null</code>
	 */
	public static int hcode(Object o) {
		if (o == null) {
			return 19;
		}
		return o.hashCode();
	}

	/**
	 * Creates a (possibly inaccurate) representation of the graph that can be read by Dot.
	 * 
	 * Note that the <code>toString()</code> method of each node and edge label is used
	 * to supply the name of the node or edge in Dot. If the name contains
	 * non-alphanumeric characters, then they are converted to underscores.
	 * This may be a problem if, for example, there is a node whose
	 * <code>toString()</code> method returns <code>"#42"</code>
	 * and another node whose <code>toString()</code> method returns <code>"@42"</code>
	 * because then both nodes will appear as
	 * node <code>"_42"</code> in the generated Dot graph,
	 * and the edges of the two nodes will be combined as well.
	 * <p/>
	 * Also, if the element to be named is null, then
	 * {@link #NULL_NAME} is used as its label. 
	 * 
	 * @returns a String representation of the graph that can be read by Dot
	 * 
	 * @see #NULL_NAME
	 */
	public static<T> String toDot(DirectedLabeledMultiGraph<T> graph) {
      //Static method cannot refer to typed DirectedLabeledMultiGraph<T>
		StringBuffer sb = new StringBuffer();
		sb.append("digraph GraphVisualization {\n");
		if (graph != null) {
			for (Iterator<T> iter = graph.getNodes(); iter.hasNext();) {
				T node = iter.next();
				String name = removeIllegalCharsFromToString(node);
				sb.append(name);
				sb.append(" [shape=\"box\" label=\"");
				sb.append(name);
				sb.append("\"];\n");
			}
			for (Iterator<Edge<T> > iter = graph.getEdges(); iter.hasNext();) {
            Edge<T> edge = iter.next();
            T from = edge.getSource();
            T to = edge.getSink();
            Object label = edge.getLabel();
            sb.append(removeIllegalCharsFromToString(from));
            sb.append(" -> ");
            sb.append(removeIllegalCharsFromToString(to));
            sb.append(" [label=\"");
            sb.append(removeIllegalCharsFromToString(label));
            sb.append("\" arrowhead=\"open\"];\n");
			}
		}
		sb.append("}\n");
		return sb.toString();
	}

	/**
	 * The String used in a Dot graph if to name an element
	 * if the element to be named is <code>null</code>.
	 * The value of this string is <code>"NULL"</code>.
	 */
	public static final String NULL_NAME = "NULL";

	private static String removeIllegalCharsFromToString(Object o) {
		if (o == null) {
			return NULL_NAME;
		}
		return o.toString().replaceAll("\\W", "_");
	}

}
