package ps6.graph;

/**
 * <code>MutableDirectedLabeledMultiGraph</code> represents a directed labeled multi-graph
 * to which nodes or directed labeled edges may be added or removed.
 * <p/>
 * Some graph implementations may have restrictions on the nodes and edge labels they may contain.
 * For example, some implementations prohibit null nodes or null edge labels.
 * Attempting to insert an ineligible value throws an <code>IllegalArgumentException</code>.
 * <p/>
 * <b>Note</b>: Throughout this API, the notation <code>e1.equals(e2)</code> is used to mean
 * <code>(e1==null ? e2==null : e1.<b>equals</b>(e2))</code> where the second
 * <b>equals</b> is the method that tests for object equality.
 *
 * @see java.lang.IllegalArgumentException
 *
 * @author mbolin
 * 
 */
public interface MutableDirectedLabeledMultiGraph<T> extends DirectedLabeledMultiGraph<T> {

    /**
     * Adds <code>obj</code> as a node in the graph.
     * 
     * @effects If the graph already contains a node <code>n</code> 
     * such that <code>n.equals(obj)</code>, then <code>addNode()</code> will return false
     * and the graph will remain unchanged.
     * Otherwise, <code>obj</code> will be added as a node in the graph,
     * and <code>addNode()</code> will return true.
     * 
     * @modifies <code>nodes</code>
     * @return <code>true</code> if <code>obj</code> is added as a node in the graph,
     *   <code>false</code> otherwise
     * @throws IllegalArgumentException if some aspect of <code>obj</code> prevents
     *   it from being added to this graph
     */
    public boolean addNode(T obj);
    
    /**
     * Adds <code>edge</code> as an edge in the graph.
     * 
     * @effects If there are nodes in the graph <code>n1,n2</code> such that
     * <code>n1.equals(edge.getSource())</code> and <code>n2.equals(edge.getSink())</code>,
     * and there is no edge from <code>n1</code> to <code>n2</code>
     * with a label <code>l</code> such that <code>l.equals(edge.getLabel())</code>,
     * then an edge e will be added to the graph, such that
     * e.source == n1, e.sink == n2, e.label = edge.label
     * and <code>addEdge()</code> will return true.
     * Otherwise, the graph will remain unchanged and
     * <code>addEdge()</code> will return false.
     * 
     * @modifies <code>edges</code>
     * @return <code>true</code> if the edge is added to the graph,
     *   <code>false</code> otherwise
     * @throws IllegalArgumentException if some aspect of <code>edge</code> prevents
     *   it from being added to this graph
     */
    public boolean addEdge(Edge<T> edge);
    
    /**
     * Removes the node <code>obj</code> from the graph.
     * 
     * @effects If there does not exist a node in the graph <code>n</code>
     * such that <code>n.equals(obj)</code>, then <code>removeNode()</code>
     * will return false and the graph will remain unchanged.
     * Otherwise, the node <code>n</code> such that <code>n.equals(obj)</code>
     * will be removed from the graph, as well as all edges from the graph
     * that have <code>n</code> as a source or sink,
     * and <code>removeNode()</code> will return <code>true</code>.
     * 
     * @modifies nodes, edges
     * @return <code>true</code> if the node was removed,
     *   <code>false</code> otherwise
     */
    public boolean removeNode(T obj);
    
    /**
     * Removes the edge <code>edge</code> from the graph.
     * 
     * @effects If there are nodes in the graph <code>n1,n2</code> such that
     * <code>n1.equals(edge.getSource())</code> and <code>n2.equals(edge.getSink())</code>,
     * and there is an edge from <code> n1</code> to <code>n2</code>
     * with a label <code>l</code> such that <code>l.equals(edge.getLabel())</code>,
     * then this edge will be removed from the graph,
     * and <code>removeEdge()</code> will return false.
     * Otherwise, the graph will remain unchanged and
     * <code>removeEdge()</code> will return false.
     * 
     * @modifies edges
     * @return <code>true</code> if the edge was removed,
     *   <code>false</code> otherwise
     */
    public boolean removeEdge(Edge<T> edge);

}
