package ps6.model;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Random;
import java.util.Set;

import ps6.graph.DefaultEdge;
import ps6.graph.DefaultMutableDirectedLabeledMultiGraph;
import ps6.graph.DirectedLabeledMultiGraph;
import ps6.graph.Edge;
import ps6.graph.GraphUtils;

/**
 * <code>BoggleBoard</code> is a 4x4 Boggle board.
 * 
 * @author mbolin
 * @author cwo (modified for Java 1.5)
 */
public class BoggleBoard implements Board {

	/* 
	 * AF:
	 * 
	 *    all c, r: 0..BOARD_SIZE |
	 * 		cubeFaceStrings [r][c] =  buckets[k].cube.faces [buckets[k].face].displayLetters
	 * 		where k = c + r * BOARD_SIZE
	 * 
	 * That is, the string appearing at row r and column c is obtained by taking the
	 * appropriate CubeBucket in the buckets array, taking the face at the face index given by
	 * that CubeBucket, and taking its associated string.
	 *
	 * Explanation of rep
	 * The CubeBuckets are represented as a List so that it is convenient to
	 * shuffle the board using Collections.shuffle()  
	 * 
	 * A redundant representation as a DirectedLabeledMultiGraph is used to enable word search.
	 * Every node in the graph
	 * is a CubeBucket. Each CubeBucket has edges connecting it to
	 * its neighboring CubeBuckets, as these edges represent valid
	 * paths for connecting cubes to form words in Boggle.
	 * 
	 * The graph has an additional CubeBucket node that does not represent
	 * a CubeBucket on the board: SPECIAL_BUCKET. The SPECIAL_BUCKET
	 * node has one outgoing edge to each bucket on the board.
	 * It is convenient to start searches of the board from
	 * SPECIAL_BUCKET. 
	 * 
	 * RI:   
    *     .
    *     .
    *     .
    *     <FILL IN HERE!>
    *     .
    *     .
    *     .
	 */
	private void checkRep() {
	    if (!CHECKREP_ENABLED) {
	        return;
	    }
	}

	/**
	 * Visualizes MultiGraph generated by createCubeBucketGraph() in a .dot file
	 * because it is easier to verify that createCubeBucketGraph() did the right
	 * thing by looking at a PNG than it is to write a JUnit test.
	 */
	public static void main(String[] args) {
		Duple<DirectedLabeledMultiGraph<CubeBucket>,CubeBucket[]> ret = createCubeBucketGraph();
		DirectedLabeledMultiGraph<CubeBucket> graph = ret.first;
		String dot = GraphUtils.toDot(graph);
		System.out.println(dot);
	}

	private static final int BOARD_SIZE = 4;

	private static final int NUM_CUBES = BOARD_SIZE * BOARD_SIZE;

	private DirectedLabeledMultiGraph<CubeBucket> graph;

	private CubeBucket[] buckets;

	private final List<Cube> cubes = new ArrayList<Cube>();

	private final Random random = new Random();

	private Set<ShuffleListener> shuffleListeners = new HashSet<ShuffleListener>();

	private static final boolean CHECKREP_ENABLED = true;

	private static final String EAST = "E";

	private static final String WEST = "W";

	private static final String NORTH = "N";

	private static final String SOUTH = "S";

	private static final String NORTHEAST = "NE";

	private static final String NORTHWEST = "NW";

	private static final String SOUTHEAST = "SE";

	private static final String SOUTHWEST = "SW";

    private static final CubeBucket SPECIAL_BUCKET = new CubeBucket(-1) {
        public String getCapitalLetters() { return ""; }
    };
	
	/**
	 * Creates a MultiGraph of CubeBuckets that has the following topology:
	 * 
	 * 0  -  1  -  2  -  3
	 * |  X  |  X  |  X  |
	 * 4  -  5  -  6  -  7
	 * |  X  |  X  |  X  |
	 * 8  -  9  -  10 -  11
	 * |  X  |  X  |  X  |
	 * 12 -  13 -  14 -  15
	 * 
	 * Thus, it is a 4x4 grid of CubeBucket nodes where the buckets are indexed
	 * from 1 to 16. Each - in the diagram represents two edges: one from the
	 * left node to the right node (labeled "E") and one from the right node to
	 * the left node (labeled "W"). (All the labels correspond to compass
	 * directions: north, south, east, west.) Each | in the diagram represents
	 * two edges: one from the node above to the node below (labeled "S") and
	 * one from the node below to the node above (labeled "N"). Finally, each X
	 * in the diagram represents four diagonal edges, labeled "NE", "NW", "SE",
	 * "SW". Thus, each node is connected to its immediate neighbors, including
	 * diagonal neighbors.
	 * 
	 * There is also a special CubeBucket with index -1 that is not part of this
	 * graph (not shown in ASCII diagram). This bucket has outgoing edges from
	 * itself to the other 16 buckets, but no incoming edges. The label for each
	 * of these edges is String.valueOf(i) where i is the index of the bucket to
	 * which it points.
	 * 
	 * @return a Duple where: the first element is the MultiGraph of CubeBuckets 
    *         and the second element is a CubeBucket array a where a[i] is the 
    *         CubeBucket with index i
	 */
	private static Duple<DirectedLabeledMultiGraph<CubeBucket>,CubeBucket[]> createCubeBucketGraph() {
		DefaultMutableDirectedLabeledMultiGraph<CubeBucket> graph = new DefaultMutableDirectedLabeledMultiGraph<CubeBucket>();

		CubeBucket[] buckets = new CubeBucket[NUM_CUBES];
		for (int i = 0; i < buckets.length; i++) {
			buckets[i] = new CubeBucket(i);
			graph.addNode(buckets[i]);
		}

		graph.addNode(SPECIAL_BUCKET);

		for (int i = 0; i < buckets.length; i++) {
			graph.addEdge(new DefaultEdge<CubeBucket>(SPECIAL_BUCKET, buckets[i], String
					.valueOf(i)));

			boolean north, south, east, west;
			north = south = east = west = false;
			if (i >= BOARD_SIZE) {
				north = true;
				graph.addEdge(new DefaultEdge<CubeBucket>(buckets[i], buckets[i
						- BOARD_SIZE], NORTH));
			}
			if (i < (BOARD_SIZE * (BOARD_SIZE - 1))) {
				south = true;
				graph.addEdge(new DefaultEdge<CubeBucket>(buckets[i], buckets[i
						+ BOARD_SIZE], SOUTH));
			}
			if (i % BOARD_SIZE != 0) {
				west = true;
				graph
						.addEdge(new DefaultEdge<CubeBucket>(buckets[i], buckets[i - 1],
								WEST));
			}
			if (i % BOARD_SIZE != (BOARD_SIZE - 1)) {
				east = true;
				graph
						.addEdge(new DefaultEdge<CubeBucket>(buckets[i], buckets[i + 1],
								EAST));
			}
			if (north && west) {
				graph.addEdge(new DefaultEdge<CubeBucket>(buckets[i], buckets[i - 1
						- BOARD_SIZE], NORTHWEST));
			}
			if (north && east) {
				graph.addEdge(new DefaultEdge<CubeBucket>(buckets[i], buckets[i + 1
						- BOARD_SIZE], NORTHEAST));
			}
			if (south && west) {
				graph.addEdge(new DefaultEdge<CubeBucket>(buckets[i], buckets[i - 1
						+ BOARD_SIZE], SOUTHWEST));
			}
			if (south && east) {
				graph.addEdge(new DefaultEdge<CubeBucket>(buckets[i], buckets[i + 1
						+ BOARD_SIZE], SOUTHEAST));
			}
		}
      
      Duple<DirectedLabeledMultiGraph<CubeBucket>,CubeBucket[]> my_ret = 
            new Duple<DirectedLabeledMultiGraph<CubeBucket>,CubeBucket[]>();

      my_ret.first = graph;
      my_ret.second = buckets;

		return my_ret;
	}

	/**
	 * Creates a new, shuffled board
	 */
	public BoggleBoard() {
		Duple<DirectedLabeledMultiGraph<CubeBucket>,CubeBucket[]> ret = createCubeBucketGraph();
      this.graph = ret.first;
      this.buckets = ret.second;

		// add cubes to a List to take advantage of Collections.shuffle()
		for (Iterator<Cube> iter = Cube.allCubes(); iter.hasNext();) {
			cubes.add(iter.next());
		}

		shuffle();
		checkRep();
	}

	/**
	 * Creates a new board using the specified buckets.
	 * This constructor should only be used within the ps6.model package
	 * to create artifical boards used for JUnit testing.
	 */
	BoggleBoard(CubeBucket[] specificBuckets) {
		Duple<DirectedLabeledMultiGraph<CubeBucket>,CubeBucket[]> ret = createCubeBucketGraph();
      this.graph = ret.first;
      this.buckets = ret.second;

		for (int i = 0; i < buckets.length; i++) {
			buckets[i].setCube(specificBuckets[i].getCube());
			buckets[i].setFace(specificBuckets[i].getFace());
		}
	}
	
	/** @{inheritDoc} */
	public int getBoardSize() {
		checkRep();
		return BOARD_SIZE;
	}

	/** @{inheritDoc} */
	public void shuffle() {
		checkRep();
		try {
			Collections.shuffle(cubes);
			int index = 0;
			for (Iterator<Cube> iter = cubes.iterator(); iter.hasNext();) {
				Cube cube = iter.next();
				CubeBucket cb = buckets[index++];
				cb.setCube(cube);
				cb.setFace(random.nextInt(Cube.NUM_FACES));
			}
			fireShuffleEvent();
		} finally {
			checkRep();
		}
	}

	/** @{inheritDoc} */
	public String getLettersAt(int row, int column) {
		checkRep();
		try {
			if (row < 0) {
				throw new IllegalArgumentException("row cannot be negative");
			} else if (column < 0) {
				throw new IllegalArgumentException("column cannot be negative");
			} else if (row >= BOARD_SIZE) {
				throw new IllegalArgumentException("row cannot exceed "
						+ BOARD_SIZE);
			} else if (column >= BOARD_SIZE) {
				throw new IllegalArgumentException("column cannot exceed "
						+ BOARD_SIZE);
			}
			return buckets[column + row * BOARD_SIZE].getDisplayLetters();
		} finally {
			checkRep();
		}
	}

	/** @{inheritDoc} */
	public boolean containsWord(String word) {
		checkRep();
		try {
			if (word == null) {
				return false;
			}
			return containsWord(word.toUpperCase(), SPECIAL_BUCKET, new HashSet<CubeBucket>());
		} finally {
			checkRep();
		}
	}
	
	private boolean containsWord(String word, CubeBucket toExpand, Set<CubeBucket> visitedBuckets) {
	    if (visitedBuckets.contains(toExpand)) {
	        return false;
	    }
	    if (word.startsWith(toExpand.getCapitalLetters())) {
	        if (word.length() == toExpand.getCapitalLetters().length()) {
	            return true;
	        }
	        String suffix = word.substring(toExpand.getCapitalLetters().length());
	        for (Iterator<Edge<CubeBucket> > edges = graph.getOutgoingEdges(toExpand); edges.hasNext(); ) {
	            Edge<CubeBucket> edge = edges.next();
	            CubeBucket bucket = edge.getSink();
	            Set<CubeBucket> visited = new HashSet<CubeBucket>(visitedBuckets);
	            visited.add(toExpand);
	            boolean result = containsWord(suffix, bucket, visited);
	            if (result) {
	                return true;
	            }
	        }
	        return false;
	    }
	    return false;
	}

	/** @{inheritDoc} */
	public boolean addShuffleListener(ShuffleListener l) {
		checkRep();
		try {
			if (l == null) {
				throw new IllegalArgumentException("shuffle listener was null");
			}
			return shuffleListeners.add(l);
		} finally {
			checkRep();
		}
	}

	/** @{inheritDoc} */
	public boolean removeShuffleListener(ShuffleListener l) {
		checkRep();
		try {
			return shuffleListeners.remove(l);
		} finally {
			checkRep();
		}
	}

	/**
	 * @effects Invokes the {@link ShuffleListener#boardShuffled(Board)} method
	 * for all ShuffleListeners of this board 
	 */
	private void fireShuffleEvent() {
		checkRep();
		try {
			for (Iterator iter = shuffleListeners.iterator(); iter.hasNext();) {
				ShuffleListener listener = (ShuffleListener) iter.next();
				listener.boardShuffled(this);
			}
		} finally {
			checkRep();
		}
	}
	
	/**
	 * @returns a 4-line ASCII representation of the board, such as:
	 * <pre>
	 * N S P U
     * E O T R
     * W F I E
     * S P A P
	 * </pre>
	 */
	public String toString() {
		StringBuffer sb = new StringBuffer();
		for (int row = 0; row < BOARD_SIZE; row++) {
			for (int col = 0; col < BOARD_SIZE; col++) {
				sb.append(getLettersAt(row, col));
				sb.append(" ");
			}
			if (row < BOARD_SIZE - 1) {
			    sb.append("\n");
			}
		}
		return sb.toString();
	}
	
}
